/* 3D BSP tree maker
 * Ben Lin 27/11/98
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stdarg.h>
#include <kb.h>
#include "blin.h"
#include "yuck.h"
#include "io.h"

#define epsilon 1e-13

#define frontclip 0.5

#define angleinc M_PI / 128

#define D0 1	// For use in diffuse reflection formula
#define SWIDTH 320
#define SHEIGHT 200
#define SMIDX 160
#define SMIDY 100

#define VecList(V) V[0], V[1], V[2]
#define VecDot(A, B) ((A[0]*B[0]) + (A[1]*B[1]) + (A[2]*B[2]))
#define VecNorm(V) sqrt(VecDot(V, V))
#define VecScalar(V, assign_op, k) \
(	V[0] assign_op k, \
	V[1] assign_op k, \
	V[2] assign_op k \
)
#define VecVec(A, assign_op, B) \
(	A[0] assign_op B[0], \
	A[1] assign_op B[1], \
	A[2] assign_op B[2] \
)
#define VecVecVec(A, assign_op, B, op, C) \
(	A[0] assign_op B[0] op C[0], \
	A[1] assign_op B[1] op C[1], \
	A[2] assign_op B[2] op C[2]\
)

#define VecCross(A, B, C) \
(	A[0] = B[1]*C[2] - B[2]*C[1], \
	A[1] = B[2]*C[0] - B[0]*C[2], \
	A[2] = B[0]*C[1] - B[1]*C[0]\
)

#define VecList(V) V[0], V[1], V[2]

#if 0
#define VecNorm2(V) sqrt(VecDot2(V, V))
#define VecList2(V) V[0], V[1]
#define VecDot2(A, B) ((A[0]*B[0]) + (A[1]*B[1]))

#define VecScalar2(V, assign_op, k) \
(	V[0] assign_op k, \
	V[1] assign_op k \
)
#define VecVec2(A, assign_op, B) \
(	A[0] assign_op B[0], \
	A[1] assign_op B[1] \
)
#define VecVecVec2(A, assign_op, B, op, C) \
(	A[0] assign_op B[0] op C[0], \
	A[1] assign_op B[1] op C[1] \
)
#define MatVec2(A, M, B) \
(	A[0] = M[0][0]*B[0] + M[0][1]*B[1], \
	A[1] = M[1][0]*B[0] + M[1][1]*B[1] \
)
#endif

#define stepsize 0.5

typedef double vector[3];
typedef vector matrix[3];
typedef int coord[2];

enum {
	t_back = -1,
	t_front = 1,
	t_on = 0,

	t_cross = 2,
	/* t_cross_plus means it crosses, and the wall start is in front,
	 * wall end is behind, and t_cross_minus is the other way around */
	t_cross_plus = 2,
	t_cross_minus = -2,
	string_max = 80,
	debug = 0
};

struct vlelem {
	struct vlelem *next;
	struct vlelem *prev;
	vector v;
};

typedef struct vlelem vlelem;

struct face_s {
	int ecount;

	// TODO: take this out, and make an origface struct
	int (*e)[2];

	vlelem *p;	// point coords, in doubly-linked list

	vector n;	//	normal
	double colour[3];
	double light[3];
	double diffuse, spec;
	int specpower;
};

typedef struct face_s face_s;

struct object_s {
	int vcount;			// no. of vertices
	int fcount;			// no. of faces
	face_s *face;
};

typedef struct object_s object_s;

struct flelem {
	struct flelem *next;
	face_s *face;
};

typedef struct flelem flelem;

struct onedge_s {
	//face_s *face;
	vlelem *ve;
};

typedef struct onedge_s onedge_s;

struct plane {
	vector p;
	vector n;
};

typedef struct plane plane;

struct bsptreenode {
	struct bsptreenode *front;
	struct bsptreenode *back;
	plane cut;
	flelem *fl;
};

typedef struct bsptreenode bsptreenode;

struct ilist {
	struct ilist *next;
	int x;
};

typedef struct ilist ilist;

struct light_s {
	vector loc;	// location
	double colour[3];
};

typedef struct light_s light_s;

struct sceneinfo {
	vector eye;
	vector at;
	vector up;
	vector right;
	vector amblight;
	double focal;
	double scale;
	double wwidth;
	double wheight;
	int swidth;
	int sheight;
} sc;

vector *vertex;
vector *svertex;
coord *scoord;
light_s *light;
object_s *object;
int lightcount, objectcount, vcount;
bsptreenode *bsproot;
onedge_s *onedge;
int onedgecount;
char img[SWIDTH*SHEIGHT];
ilist inter[SHEIGHT];

// for my own malloc routines
// (had trouble with standard ones - memory leak somewhere)
ilist *ipool;
int ipoolcount = 0;

double sinangleinc, cosangleinc;
FILE *bspfile;
char bspfilemode = 0;

//void split(face_s *f, plane *cut, flelem *above, flelem *on, flelem *below);

// prototypes
void loadbsptree(bsptreenode *node);
void savebsptree(bsptreenode *node);
char splitcut(face_s *face, plane *cut);
void choosebest(flelem *fl, plane *cut);
void drawscene(void);
void key_movement(void);
void drawface(const face_s *face);
void painter(const bsptreenode *node);
inline void plot(int x, int y, byte c);
void bresenham(int x1, int y1, int x2, int y2, byte c);
void makebsptree(bsptreenode *node, flelem *fl);
char classify(face_s *face, plane *cut);
void loadfile(char *filename);

int main(int argc, char **argv) {
	int i;
	char quitflag;

	if (argc < 2) {
		fprintf(stderr, "No filename given\n");
		exit(1);
	}

	if (argc == 3) {
		bspfilemode = 1;
		bspfile = fopen(argv[2], "r");

		if (!bspfile) {
			bspfilemode = 2;
			bspfile = fopen(argv[2], "w");
			if (!bspfile) {
				perror(argv[2]);
				exit(3);
			}
		}
	}

	loadfile(argv[1]);

	initgraphics();

	cosangleinc = cos(angleinc);
	sinangleinc = sin(angleinc);

	// Initialize ilists
	for (i=0; i<SHEIGHT; i++) inter[i].next = NULL;

	drawimage(img);

	kb_install(0);
	quitflag = 0;

	// Main loop
	while(!quitflag) {
		drawscene();
		kb_update();
		key_movement();
		if (kb_key(KB_SCAN_Q)) quitflag = 1;
	}

	settextmode();
	return(0);
}

void drawscene(void) {
	int i;

	for (i=0; i<SWIDTH*SHEIGHT; i++) img[i] = 0;

	painter(bsproot);
	drawimage(img);
}

/* Parse a .gen file */

void loadfile(char *filename) {
	int i, j, k;
	int vmax;
	int v;
	int n;
	double f;
	int *fromv, *tov;
	object_s *o;
	FILE *fp;
	flelem *fltail, *flhead;
	face_s *face;
	vlelem *ve;
	vector p;

	flhead = fltail = NEW(flelem);
	flhead->next = NULL;

	if (!(fp = fopen(filename, "r"))) {
		perror(filename);
		exit(1);
	}

	currentfile(fp);

	readin("%lf", &sc.eye[0], "eye x");
	readin("%lf", &sc.eye[1], "eye y");
	readin("%lf", &sc.eye[2], "eye z");
	readin("%lf", &p[0], "view at x");
	readin("%lf", &p[1], "view at y");
	readin("%lf", &p[2], "view at z");
	VecVecVec(sc.at, =, p, -, sc.eye);
	f = VecNorm(sc.at);
	VecScalar(sc.at, /=, f);
	readin("%lf", &p[0], "view up x");
	readin("%lf", &p[1], "view up y");
	readin("%lf", &p[2], "view up z");

	f = VecDot(p, sc.at);
	VecVecVec(sc.up, =, p, -, f*sc.at);
	f = VecNorm(sc.up);
	VecScalar(sc.up, /=, f);

	VecCross(sc.right, sc.at, sc.up);

	readin("%lf", &sc.focal, "focal length");
	readin("%lf", &sc.wwidth, "window width");
	readin("%lf", &sc.wheight, "window height");

	readin("%d", &sc.swidth, "screen height");
	readin("%d", &sc.sheight, "screen height");
	sc.swidth = SWIDTH;
	sc.sheight = SHEIGHT;

	sc.scale = sc.swidth / sc.wwidth;

	readin("%lf", &sc.amblight[0], "ambient light r");
	readin("%lf", &sc.amblight[1], "ambient light g");
	readin("%lf", &sc.amblight[2], "ambient light b");

	readin("%d", &lightcount, "number of lights");

	light = NEWARRAY(light_s, lightcount);

	for (i=0; i<lightcount; i++) {
		readin("%lf", &light[i].loc[0], "light #%d x", i);
		readin("%lf", &light[i].loc[1], "light #%d y", i);
		readin("%lf", &light[i].loc[2], "light #%d z", i);

		readin("%lf", &light[i].colour[0], "light #%d r", i);
		readin("%lf", &light[i].colour[1], "light #%d g", i);
		readin("%lf", &light[i].colour[2], "light #%d b", i);
	}

	readin("%d", &objectcount, "number of objects");
	o = NEWARRAY(object_s, objectcount);
	object = o;

	readin("%d", &vcount, "total vertices");
	vertex = NEWARRAY(vector, vcount);
	
	fromv = NEWARRAY(int, vcount);
	tov = NEWARRAY(int, vcount);

	vmax = 0;
	v = 0;

	for (i=0; i<objectcount; i++) {

		readin("%d", &o[i].vcount, "number of vertices in object");

		if (vmax < o[i].vcount) vmax = o[i].vcount;

		for (j=0; j<o[i].vcount; j++) {
			readin("%lf", &vertex[v][0], "vertex #%d x", j);
			readin("%lf", &vertex[v][1], "vertex #%d y", j);
			readin("%lf", &vertex[v][2], "vertex #%d z", j);
			v++;
		}

		readin("%d", &o[i].fcount, "number of faces");
		o[i].face = NEWARRAY(face_s, o[i].fcount);
		
		printf("object %d\n", i);
		for (j=0; j<o[i].fcount; j++) {
			//printf("\tface %d\n", j);

			k = 0;
			do {
				readin("%d", &n, "edge's from-vertex");

				if (n != -1) {
					fromv[k] = v - o[i].vcount + n;
					readin("%d", &n, "edge's to-vertex");
					tov[k] = v - o[i].vcount + n;
					k++;
				}
			} while (n != -1);
			o[i].face[j].ecount = k;
			o[i].face[j].e = (int (*)[2]) xmalloc(sizeof(int) * 2 * o[i].face[j].ecount);

			/* Renumber consistently */
			o[i].face[j].e[0][0] = fromv[0];
			o[i].face[j].e[0][1] = tov[0];
			fromv[0] = tov[0] = -1;
			for (k=1; k<o[i].face[j].ecount; k++) {
				n = 1;
				while (1) {
					if (fromv[n] == o[i].face[j].e[k - 1][1]) {
						o[i].face[j].e[k][0] = fromv[n];
						o[i].face[j].e[k][1] = tov[n];
						fromv[k] = tov[k] = -1;
						break;
					} else if (tov[n] == o[i].face[j].e[k - 1][1]) {
						o[i].face[j].e[k][0] = tov[n];
						o[i].face[j].e[k][1] = fromv[n];
						fromv[k] = tov[k] = -1;
						break;
					}
				 	n++;
					if (n > o[i].face[j].ecount) {
						fprintf(stderr, "face not closed\n");
						exit(1);
					}
				}

			}

			readin("%lf", &o[i].face[j].n[0], "normal x");
			readin("%lf", &o[i].face[j].n[1], "normal y");
			readin("%lf", &o[i].face[j].n[2], "normal z");

			/* Colour */
			readin("%lf", &o[i].face[j].colour[0], "face r");
			readin("%lf", &o[i].face[j].colour[1], "face g");
			readin("%lf", &o[i].face[j].colour[2], "face b");

			/* Diffuse, spec(ular), spec. power */
			readin("%lf", &o[i].face[j].diffuse, "diffuse");
			readin("%lf", &o[i].face[j].spec, "specular");
			readin("%d", &o[i].face[j].specpower, "specular power");

			/* Work out all we can about the light it gives off */
			VecVecVec(o[i].face[j].light, =, o[i].face[j].colour, *, sc.amblight);
			for (k=0; k<lightcount; k++) {
				// TODO: calculate centroid and use that instead
				VecVecVec(p, =, light[k].loc, -, vertex[o[i].face[j].e[0][0]]);
				f = VecNorm(p);
				VecScalar(p, /=, f);
				f = VecDot(p, o[i].face[j].n);
				VecVec(o[i].face[j].light, +=,
						o[i].face[j].diffuse * f * o[i].face[j].colour);
			}

			// copy original face to new face, represent in
			// doubly-linked list
			face = NEW(face_s);
			*face = o[i].face[j];
			ve = face->p = NEW(vlelem);
			VecVec(ve->v, =, vertex[face->e[0][0]]);
			for (k=1; k<face->ecount; k++) {
				ve->next = NEW(vlelem);
				ve->next->prev = ve;
				ve = ve->next;
				VecVec(ve->v, =, vertex[face->e[k][0]]);
			}

			// close off the list in a circle
			ve->next = face->p;
			face->p = ve;

			// append to list of faces
			fltail->next = NEW(flelem);
			fltail = fltail->next;
			fltail->face = face;
			fltail->next = NULL;

		}

	}

	fclose(fp);

	ipool = NEWARRAY(ilist, SHEIGHT * vmax);
	vmax *= 2;
	bsproot = NEW(bsptreenode);
	onedge = NEWARRAY(onedge_s, vmax);
	if (bspfilemode == 1) {
		currentfile(bspfile);
		loadbsptree(bsproot);
		fclose(bspfile);
	} else {
		makebsptree(bsproot, flhead);
	}

	if (bspfilemode == 2) {
		savebsptree(bsproot);
	}
	
	scoord = NEWARRAY(coord, vmax);
	svertex = NEWARRAY(vector, vmax);

	xfree(fromv);
	xfree(tov);

}

void makebsptree(bsptreenode *node, flelem *fl) {
	flelem *flfronthead, *flbackhead;
	flelem *fefront, *feback, *feon, *fe;
	flelem *newfe;
	double d, mu;
	vector p, dir, v;
	char type;

	fefront = flfronthead = NEW(flelem);
	feback = flbackhead = NEW(flelem);
	fefront->next = feback->next = NULL;
	node->fl = NEW(flelem);
	node->fl->next = NULL;

	feon = node->fl;

	fe = fl->next;
	choosebest(fl, &node->cut);
	//VecVec(node->cut.p, =, fe->face->p->v);
	//VecVec(node->cut.n, =, fe->face->n);

	// Add this face to the on list: wrong now that I'm using choosebest
	/*
	feon->next = fe;
	feon = fe;
	fe = fe->next;
	feon->next = NULL;
	*/

	while(fe) {
		type = classify(fe->face, &node->cut);

		switch(type) {
			case t_on:
				feon->next = fe;
				feon = fe;
				fe = fe->next;
				feon->next = NULL;
				break;

			case t_front:
				fefront->next = fe;
				fefront = fe;
				fe = fe->next;
				fefront->next = NULL;
				break;
				
			case t_back:
				feback->next = fe;
				feback = fe;
				fe = fe->next;
				feback->next = NULL;
				break;

			default:
				//fprintf(stderr, "Warning: crossed not handled\n");
				if (0) {
					vlelem *ve;
					int i;

					printf("cut->p: %.2lf %.2lf %.2lf\n", VecList(node->cut.p));
					printf("cut->n: %.2lf %.2lf %.2lf\n", VecList(node->cut.n));
					ve = fe->face->p;
					printf("face vertices:\n");
					do {
						printf("%.2lf %.2lf %.2lf\n", VecList(ve->v));
						ve = ve->next;
					} while (ve != fe->face->p);
					printf("onedge:\n");
					for (i=0; i<onedgecount; i++) {
						printf("%.2lf %.2lf %.2lf\n", VecList(onedge[i].ve->v));
					}
					printf("\n");
				}
				if (onedgecount != 2) {
					printf("Warning: non-convex polygon split\n");
					printf("(%d)\n", onedgecount);
					fe = fe->next;
				} else {
					vlelem *ve;

					newfe = NEW(flelem);
					newfe->face = NEW(face_s);
					*newfe->face = *fe->face;

					newfe->face->p = NEW(vlelem);
					VecVec(newfe->face->p->v, =, onedge[0].ve->v);
					newfe->face->p->next = onedge[0].ve->next;
					newfe->face->p->next->prev = newfe->face->p;

					newfe->face->p->prev = ve =  NEW(vlelem);
					VecVec(ve->v, =, onedge[1].ve->v);
					ve->prev = onedge[1].ve->prev;
					ve->next = newfe->face->p;
					ve->prev->next = ve;

					onedge[0].ve->next = onedge[1].ve;
					onedge[1].ve->prev = onedge[0].ve;

					//printf("poly 1 ");
					ve = fe->face->p;
					fe->face->ecount = 0;
					//printf("face vertices:\n");
					do {
						//printf("%.2lf %.2lf %.2lf\n", VecList(ve->v));
						fe->face->ecount++;
						ve = ve->next;
					} while (ve != fe->face->p);

					//printf("poly 2 ");
					ve = newfe->face->p;
					newfe->face->ecount = 0;
					//printf("face vertices:\n");
					do {
						//printf("%.2lf %.2lf %.2lf\n", VecList(ve->v));
						newfe->face->ecount++;
						ve = ve->next;
					} while (ve != newfe->face->p);

					type = classify(fe->face, &node->cut);
					if (type == t_front) {
						fefront = fefront->next = fe;
						fe = fe->next;
						fefront->next = NULL;

						feback = feback->next = newfe;
						feback->next = NULL;
					} else {
						feback = feback->next = fe;
						fe = fe->next;
						feback->next = NULL;

						fefront = fefront->next = newfe;
						fefront->next = NULL;
					}
				}

				break;
		}
	}

	// recurse on front and back nodes
	if (flfronthead->next) {
		node->front = NEW(bsptreenode);
		makebsptree(node->front, flfronthead);
	} else {
		node->front = NULL;
	}
	if (flbackhead->next) {
		node->back = NEW(bsptreenode);
		makebsptree(node->back, flbackhead);
	} else {
		node->back = NULL;
	}

}

char classify(face_s *face, plane *cut) {
	char type;
	vlelem *ve, *newve;
	vector v, dir;
	double d, d2, mu;
	char sign;

	ve = face->p;
	onedgecount = 0;
	type = t_on;

	do {
		VecVecVec(v, =, ve->v, -, cut->p);
		d = VecDot(cut->n, v);

		if (d > epsilon) {
			sign = 1;
			if (type == t_on) type = t_front;
			else if (type == t_back) type = t_cross;
		} else if (d < -epsilon) {
			sign = -1;
			if (type == t_on) type = t_back;
			if (type == t_front) type = t_cross;
		} else sign = 0;

		VecVecVec(v, =, ve->next->v, -, cut->p);
		d2 = VecDot(cut->n, v);

		if (d2 > epsilon) {
		} else if (d2 < -epsilon) {
			sign = -sign;
		} else sign = 1;

		if (sign < 0) {
			// it crosses
			// find point of intersection
			VecVecVec(dir, =, ve->next->v, -, ve->v);
			d = VecDot(cut->n, dir);
			VecVecVec(v, =, cut->p, -, ve->v);
			mu = VecDot(cut->n, v) / d;

			// insert edge
			VecVecVec(v, =, ve->v, +, mu * dir);

			//VecVec(v, -=, cut->p);
			//printf("should be 0: %lf\n", VecDot(cut->n, v));

			newve = NEW(vlelem);
			VecVec(newve->v, =, v);
			newve->next = ve->next;
			ve->next = newve;
			newve->next->prev = newve;
			newve->prev = ve;

			onedge[onedgecount].ve = newve;
			onedgecount++;

			ve = ve->next;
		} else if (sign == 0) {
			// otherwise from vertex lies on the plane
			// add from vertex to onedge list

			onedge[onedgecount].ve = ve;
			onedgecount++;
		}

		ve = ve->next;
	} while(ve != face->p);

	return(type);
}

void painter(const bsptreenode *node) {
	flelem *fe;
	vector p;
	double d;

	VecVecVec(p, =, sc.eye, -, node->cut.p);
	d = VecDot(node->cut.n, p);

	if (d > epsilon) {
		if (node->back) {
			painter(node->back);
		}

		fe = node->fl->next;
		while(fe) {
			drawface(fe->face);
			fe = fe->next;
		}

		if (node->front) {
			painter(node->front);
		}
	} else if (d < epsilon) {
		if (node->front) {
			painter(node->front);
		}
		fe = node->fl->next;
		while(fe) {
			drawface(fe->face);
			fe = fe->next;
		}
		if (node->back) {
			painter(node->back);
		}
	}
}

void drawface(const face_s *face) {
	int i, n;
	int y, y2, dy, x, x2, dx;
	vlelem *ve;
	vector v, v2;
	int ymin, ymax;
	double d;
	ilist *ip, *itemp;
	byte colour;

	n = 0;
	ve = face->p;

	do {
		// Transform each vertex appropriately
		VecVecVec(v, =, ve->v, -, sc.eye);
		svertex[n][0] = VecDot(v, sc.right);
		svertex[n][1] = VecDot(v, sc.up);
		svertex[n][2] = VecDot(v, sc.at);

		if (svertex[n][2] < frontclip) return;

		//printf("d1: %lf %lf %lf\n", ve->v[0], ve->v[1], ve->v[2]);
		//printf("d2: %lf %lf %lf\n", svertex[n][0], svertex[n][1], svertex[n][2]);
		d = sc.scale * sc.focal / svertex[n][2];

		scoord[n][0] = SMIDX + (int) (svertex[n][0] * d);
		scoord[n][1] = SMIDY - (int) (svertex[n][1] * d);
		n++;
		ve = ve->next;
	} while(ve != face->p);

	logprintf("%d: %d %d\n", 0, scoord[0][0], scoord[0][1]);
	for (i=1; i<n; i++) {
		logprintf("%d: %d %d\n", i, scoord[i][0], scoord[i][1]);
		//bresenham(scoord[i-1][0], scoord[i-1][1], scoord[i][0], scoord[i][1], 15);
	}
	//bresenham(scoord[n-1][0], scoord[n-1][1], scoord[0][0], scoord[0][1], 15);

	{
		byte rgb[3];

		d = fabs(VecDot(face->n, sc.at));
		rgb[0] = d * 63 * face->colour[0];
		rgb[1] = d * 63 * face->colour[1];
		rgb[2] = d * 63 * face->colour[2];
		colour = rgb2c(rgb);
	}

	ymax = 0;
	ymin = SHEIGHT - 1;

	for (i=0; i<n; i++) {
		int i2, i3;
		int error;
		int xinc;
		char sign2, sign3;

		i2 = i + 1;
		if (i2 == n) i2 = 0;

		y = scoord[i][1];
		y2 = scoord[i2][1];

		if (y < 0 && y2 < 0) continue;
		if (y > SHEIGHT - 1 && y2 > SHEIGHT - 1) continue;

		if (y > y2) {
			swap(y, y2);
			sign2 = -1;
			x = scoord[i2][0];
			x2 = scoord[i][0];
		} else {
			sign2 = 1;
			x = scoord[i][0];
			x2 = scoord[i2][0];
		}
		dy = y2 - y;

		dx = x2 - x;

		//clip
		if (y < 0) {
			x += dx * (-y) / dy;
			dy += y;
			y = 0;
			dx = x2 - x;
		}

		if (y2 >= SHEIGHT) {
			x2 -= dx * (y2 - SHEIGHT) / dy;
			dy -= y2 - SHEIGHT + 1;
			y2 = SHEIGHT - 1;
			dx = x2 - x;
		}

		if (ymin > y) ymin = y;
		if (ymax < y2) ymax = y2;

		if (!dy) {
			y = scoord[i][1];
			if (y >= 0 && y < SHEIGHT) {
				x = scoord[i][0];
				x2 = scoord[i2][0];

				if (x2 >= 0 && x <= SWIDTH - 1) {
					if (x < 0) {
						x = 0;
					}
					if (x2 >= SWIDTH) {
						x2 = SWIDTH - 1;
					}
					do {
						img[y * SWIDTH + x] = colour;
						x++;
					} while(x <= x2);
				}

			}
			continue;
		}

		if (dx < 0) {
			dx = -dx;
			xinc = -1;
		} else {
			xinc = 1;
		}

		error = dx - dy / 2;
		y++;
		while (error > 0) {
			x+=xinc;
			error -= dy;
		}

		while(y<y2) {
			ip = &inter[y];
			while (ip->next && ip->next->x < x)
				ip = ip->next;

			itemp = ip->next;
			ip->next = &ipool[ipoolcount++]; //NEW(ilist);
			ip->next->x = x;
			ip->next->next = itemp;

			error += dx;
			y++;
			plot(x, y, colour);
			while (error > 0) {
				x+=xinc;
				error -= dy;
				plot(x, y, colour);
			}
		}

		i3 = i2 + 1;
		if (i3 == n) i3 = 0;

		error = scoord[i3][1] - scoord[i2][1];
		if (error > 0) {
			sign3 = 1;
		} else if (error < 0) {
			sign3 = -1;
		} else {
			i3++;
			if (i3 == n) i3 = 0;
			// I should make i4 = i3 + 1
			// cos I want scoord[i4][1] - scoord[i3][1]
			// and I've worked it out the hard way
			error = scoord[i3][1] - scoord[i2][1] - error;
			sign3 = (error > 0) ? 1 : -1;
		}

		x = scoord[i][0];
		y = scoord[i][1];
		if (y >= 0 && y < SHEIGHT) {
			ip = &inter[y];
			while (ip->next && ip->next->x < x) ip = ip->next;
			itemp = ip->next;
			ip->next = &ipool[ipoolcount++]; //NEW(ilist);
			ip->next->x = x;
			ip->next->next = itemp;
		}
		if (sign3 != sign2) {
			x = scoord[i2][0];
			y = scoord[i2][1];
			if (y >= 0 && y < SHEIGHT) {
				ip = &inter[y];
				while (ip->next && ip->next->x < x) ip = ip->next;
				itemp = ip->next;
				ip->next = &ipool[ipoolcount++]; //NEW(ilist);
				ip->next->x = x;
				ip->next->next = itemp;
			}
		}

	}

	for (y=ymin, n=y*SWIDTH; y<=ymax; y++, n+=SWIDTH) {
		ip = inter[y].next;
		while (ip) {
			x = ip->x;
			itemp = ip;
			ip = ip->next;

			if (!ip) continue;

			x2 = ip->x;
			//xfree(itemp);
			itemp = ip;
			ip = ip->next;
			//xfree(itemp);
			if (x2 >= 0 && x <= SWIDTH - 1) {
				if (x < 0) {
					x = 0;
				}
				if (x2 >= SWIDTH) {
					x2 = SWIDTH - 1;
				}
				y2 = x + n;
				do {
					img[y2] = colour;
					x++;
					y2++;
				} while(x <= x2);
			}
		}
		inter[y].next = NULL;
	}
	ipoolcount = 0;

}

void bresenham(int x1, int y1, int x2, int y2, byte c) {
	int dx, dy, x, y, xinc, error;
	int i;
	
	if (y2 < y1) {
		y = y2;
		y2 = y1;
		y1 = y;
		x = x2;
		x2 = x1;
		x1 = x;
	}

	if (x2 > x1) xinc = 1;
	else if (x2 < x1) xinc = -1;
	else {
		for (y = y1; y <= y2; y++) {
			plot(x1, y, c);
		}
		return;
	}
	
	dx = x2 - x1;
	dy = y2 - y1;
	if (dx < 0) dx = -dx;
	error = -dx / 2;
	x = x1;
	y = y1;

	if (!dy) {
		for (i=0; i<=dx; i++) {
			plot(x, y, c);
			x += xinc;
		}
		return;
	}

	for (i=0; i<=dx; i++) {
		plot(x, y, c);
		x += xinc;
		error += dy;
		while (error > 0 && y < y2) {
			y++;
			plot(x, y, c);
			error -= dx;
		}
	}

}

inline void plot(int x, int y, byte c) {
	if (x < 0 || x >= SWIDTH || y < 0 || y >= SHEIGHT) return;
	img[y * SWIDTH + x] = c;
}

void key_movement(void) {
	vector v;
	double x, y;
	vector dst;

	VecVec(dst, =, sc.eye);

	if (kb_key(KB_SCAN_7_PAD) || kb_key(KB_SCAN_HOME)) {
		VecVec(sc.right, =, cosangleinc * sc.right);
		VecVec(sc.right, -=, sinangleinc * sc.up);
		VecCross(sc.up, sc.right, sc.at);
	} else if (kb_key(KB_SCAN_1_PAD) || kb_key(KB_SCAN_END)) {
		VecVec(sc.right, =, cosangleinc * sc.right);
		VecVec(sc.right, +=, sinangleinc * sc.up);
		VecCross(sc.up, sc.right, sc.at);
	}

	if (kb_key(KB_SCAN_9_PAD) || kb_key(KB_SCAN_PGUP)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			VecVec(dst, +=, stepsize * sc.up);
		} else {
			VecVec(sc.at, =, cosangleinc * sc.at);
			VecVec(sc.at, +=, sinangleinc * sc.up);
			VecCross(sc.up, sc.right, sc.at);
		}
	} else if (kb_key(KB_SCAN_3_PAD) || kb_key(KB_SCAN_PGDN)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			VecVec(dst, -=, stepsize * sc.up);
		} else {
			VecVec(sc.at, =, cosangleinc * sc.at);
			VecVec(sc.at, -=, sinangleinc * sc.up);
			VecCross(sc.up, sc.right, sc.at);
		}
	}

	if (kb_key(KB_SCAN_4_PAD) || kb_key(KB_SCAN_LEFT)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			VecVec(dst, -=, stepsize * sc.right);
		} else {
			VecVec(sc.at, =, cosangleinc * sc.at);
			VecVec(sc.at, -=, sinangleinc * sc.right);
			VecCross(sc.right, sc.at, sc.up);
		}
	} else if (kb_key(KB_SCAN_6_PAD) || kb_key(KB_SCAN_RIGHT)) {
		if (kb_shift() & KB_SHIFT_ANY_ALT) {
			VecVec(dst, +=, stepsize * sc.right);
		} else {
			VecVec(sc.at, =, cosangleinc * sc.at);
			VecVec(sc.at, +=, sinangleinc * sc.right);
			VecCross(sc.right, sc.at, sc.up);
		}
	}

	if (kb_key(KB_SCAN_8_PAD) || kb_key(KB_SCAN_UP)) {
		VecVec(dst, +=, stepsize * sc.at);
	} else if (kb_key(KB_SCAN_2_PAD) || kb_key(KB_SCAN_DOWN)) {
		VecVec(dst, -=, stepsize * sc.at);
	}

	VecVec(sc.eye, =, dst);
}

void choosebest(flelem *fl, plane *cut) {
	flelem *fe, *fe2;
	plane best;
	int splitcount, min;
	int i;

	min = 1000000;
	i = 0;
	fe = fl->next;
	while (fe && i<200) {
		VecVec(cut->p, =, fe->face->p->v);
		VecVec(cut->n, =, fe->face->n);

		splitcount = 0;
		fe2 = fl->next;
		while (fe2 && splitcount < min) {
			if (splitcut(fe2->face, cut)) {
				splitcount++;
			}
			fe2 = fe2->next;
		}

		if (min > splitcount) {
			min = splitcount;
			best = *cut;
		}

		fe = fe->next;
		i++;
	}

	*cut = best;

}

char splitcut(face_s *face, plane *cut) {
	vlelem *ve;
	vector v, dir;
	double d, d2, mu;
	char sign;

	ve = face->p;

	do {
		VecVecVec(v, =, ve->v, -, cut->p);
		d = VecDot(cut->n, v);

		if (d > epsilon) {
			sign = 1;
		} else if (d < -epsilon) {
			sign = -1;
		} else sign = 0;

		VecVecVec(v, =, ve->next->v, -, cut->p);
		d2 = VecDot(cut->n, v);

		if (d2 > epsilon) {
		} else if (d2 < -epsilon) {
			sign = -sign;
		} else sign = 1;

		if (sign < 0) {
			return(1);
		}

		ve = ve->next;
	} while(ve != face->p);

	return(0);
}

void savebsptree(bsptreenode *node) {
	flelem *fe;
	vlelem *ve;
	int i;

	fe = node->fl->next;

	while(fe) {
		fprintf(bspfile, "\n%d\n", fe->face->ecount);

		ve = fe->face->p;
		for (i=0; i<fe->face->ecount; i++) {
			fprintf(bspfile, "%lf %lf %lf\n", VecList(ve->v));
			ve = ve->next;
		}
		if (ve != fe->face->p) {
			fprintf(bspfile, "this file has errors\n");
		}
		fprintf(bspfile, "\n");

		fprintf(bspfile, "%lf %lf %lf\n", VecList(fe->face->n));
		fprintf(bspfile, "%lf %lf %lf\n", VecList(fe->face->colour));
		fprintf(bspfile, "%lf %lf %d\n", fe->face->diffuse, fe->face->spec,
				fe->face->specpower);
		
		fe = fe->next;
	}

	fprintf(bspfile, "0\n");

	fprintf(bspfile, "// cut plane, point and normal\n");
	fprintf(bspfile, "%lf %lf %lf\n", VecList(node->cut.p));
	fprintf(bspfile, "%lf %lf %lf\n", VecList(node->cut.n));

	fprintf(bspfile, "\n");
	fprintf(bspfile, "// front node\n");
	if (node->front) {
		fprintf(bspfile, "1\n");
		savebsptree(node->front);
	} else {
		fprintf(bspfile, "0\n");
	}

	fprintf(bspfile, "\n");
	fprintf(bspfile, "// back node\n");
	if (node->back) {
		fprintf(bspfile, "1\n");
		savebsptree(node->back);
	} else {
		fprintf(bspfile, "0\n");
	}

}

void loadbsptree(bsptreenode *node) {
	flelem *fe;
	vlelem *ve;
	int i, n;

	fe = node->fl = NEW(flelem);

	for(;;) {
		readin("%d", &n, "edge count");

		if (!n) break;
		
		fe = fe->next = NEW(flelem);
		fe->next = NULL;
		fe->face = NEW(face_s);

		fe->face->ecount = n;

		ve = fe->face->p = NEW(vlelem);
		for (i=0; i<n; i++) {
			readin("%lf", &ve->v[0], "vector x");
			readin("%lf", &ve->v[1], "vector y");
			readin("%lf", &ve->v[2], "vector z");
			if (i < n-1) {
				ve->next = NEW(vlelem);
				ve->next->prev = ve;
				ve = ve->next;
			} else {
				ve->next = fe->face->p;
				fe->face->p->prev = ve;
			}
		}

		readin("%lf", &fe->face->n[0], "normal x");
		readin("%lf", &fe->face->n[1], "normal y");
		readin("%lf", &fe->face->n[2], "normal z");
		readin("%lf", &fe->face->colour[0], "colour r");
		readin("%lf", &fe->face->colour[1], "colour g");
		readin("%lf", &fe->face->colour[2], "colour b");
		readin("%lf", &fe->face->diffuse, "diffuse");
		readin("%lf", &fe->face->spec, "specular");
		readin("%d", &fe->face->specpower, "specular power");
	}

	readin("%lf", &node->cut.p[0], "cut point x");
	readin("%lf", &node->cut.p[1], "cut point y");
	readin("%lf", &node->cut.p[2], "cut point z");
	readin("%lf", &node->cut.n[0], "cut normal x");
	readin("%lf", &node->cut.n[1], "cut normal y");
	readin("%lf", &node->cut.n[2], "cut normal z");

	readin("%d", &n, "front flag");
	if (n) {
		node->front = NEW(bsptreenode);
		loadbsptree(node->front);
	} else {
		node->front = NULL;
	}

	readin("%d", &n, "back flag");
	if (n) {
		node->back = NEW(bsptreenode);
		loadbsptree(node->back);
	} else {
		node->back = NULL;
	}

}
